package ConfigPatch;

use English;
use strict;
use StackTrace;
use ConfigSubs;

# This abstract class is used to patch config files. It should be
# subclassed with a class determining the exact syntax used by the
# config file to be patched.

sub keyIsUnique {
    return 0;
}

# Write @contents out to $file
# to be called after substitution
sub writeFile {
  my ($file, $contents) = @_;

  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
      $atime,$mtime,$ctime,$blksize,$blocks) = stat($file);

  my $newfile = "$file.$$";
  umask 077;
  open(OUT, ">$newfile") ||
    die "Could not write $newfile ($ERRNO)\n";
  if($EUID == 0 && defined $uid) {
      chown ($uid, $gid, $newfile) || die "Could not change owner of $newfile ($ERRNO)";
  }
  if(defined $mode) {
      chmod ($mode & 07777, $newfile) || die "Could not change mode of $newfile ($ERRNO)";
  }
  print OUT join('',@$contents) || die "Write error ($ERRNO)\n";
  close(OUT);
  (rename "$newfile", "$file") ||
    die "Could not rename $newfile to $file ($ERRNO)\n";
}


# Reads the file into an array
# TODO: separate on something else then line boundaries
sub readFile {
  my ($file) = @_;

  open(IN, "<$file") || return [];
  my @contents = <IN>;
  close(IN);
  \@contents;
}


sub closeBlock {
  my ($this, $map, $list, $seen, $out, $filter) = @_;
  if(!defined $filter) {
    $filter = sub { 1; };
  }
  foreach my $row (@$list) {
    my ($key, $desc, $accum) = @$row;
    my $compKey = 'S'.$key;
    if(!$seen->{$compKey}) {
      next unless $filter->($compKey);
      my $subFilter = $this->getDependancies($compKey, $desc);
      if(defined $subFilter) {
	$this->closeBlock($map, $list, $seen, $out, $subFilter);
      }
      push(@$out, $this->subsSetting($key, undef, $desc, $accum));
      $seen->{$compKey}=1;
    }
  }
  foreach my $key (keys %$map) {
    next if $seen->{$key};
    if($key =~ /^S(.*)$/) {
      1;
    } elsif($key =~ /^B(.*)$/) {
      # is it a block ?
      my $realKey = $1;
      next unless $filter->($key);
      $seen->{$key}=1;
      push(@$out, $map->{$key}->[3]);
      push(@$out, $this->printBlockStart($realKey,undef,$map->{$key}->[0]));
      $this->closeBlock($map->{$key}->[1],
			$map->{$key}->[2],
			{}, $out);
      push(@$out,
	   $this->printBlockEnd($realKey,undef,
				$map->{$key}->[0], $map->{$key}->[4]));
    } else {
      die "Unknown key type $key\n";
    }
  }
}

# Get dependancies
# Undef means that there are no dependancies
sub getDependancies {
  undef;
}

# Do the substitutions
sub doSubst {
  my ($this, $filename) = @_;

  my $contents = readFile($filename);
  my @out;

  my $seen={};
  my $map = $this->{map};
  my $list = $this->{list};

  my @mapStack = (); # stack of parent maps
  my @listStack = (); # stack of parent lists
  my @seenStack = (); # stack of seen maps of parents
  my @keyStack = (); # stack of block names of parents

  my $curKey = undef;  # name of current map within its parent

  $main::lineNo=0;

  # TODO: support multi-line substitutions
  foreach my $line (@$contents) {
    $main::lineNo++;
    # is this a block start ?
    my ($key, $desc) = $this->isBlockStart($line);
    if(defined $key) {
      if(defined $curKey && $this->blockAutoEnd) {
	$this->closeBlock($map, $list, $seen, \@out);
	$map = pop(@mapStack);
	$list = pop(@listStack);
	$seen = pop(@seenStack);
      } else {
	push(@keyStack, $curKey);
      }

      push(@mapStack, $map);
      push(@listStack, $list);
      push(@seenStack, $seen);

      $seen->{'B'.$key}=1;
      if(exists $map->{'B'. $key}) {
	$list = $map->{'B'. $key}->[2];
	$map = $map->{'B'. $key}->[1];
      } else {
	$map = {};
	$list = [];
      }
      $seen = { };

      $curKey = $key;
      push(@out, $line);
      next;
    }

    # is this a block end?
    my $closeDesc;
    ($key, $closeDesc) = $this->isBlockEnd($line, $curKey);
    if(defined $key) {
      if($key ne "" && $key ne $curKey) {
	die "[Subst] Misbalanced beginBlock ($curKey) and endBlock ($key)\n";
      }
      if(!defined $curKey) {
	die "[DoSubst] Attempt to close the root block\n";
      }
      $this->closeBlock($map, $list, $seen, \@out);

      $map = pop(@mapStack);
      $list = pop(@listStack);
      $seen = pop(@seenStack);
      $curKey = pop(@keyStack);

      push(@out, $line);
      next;
    }

    # is this a normal setting?
    ($key, $desc) = $this->isSetting(0, $line);
    if(defined $key) {
      my $compKey = 'S'.$key;
      if(exists $map->{$compKey}) {
	if(! $seen->{$compKey} ) {
	  # Do replacement only once
	  $line = $this->subsSetting($key, $desc, @{$map->{$compKey}});
	} elsif($this->keyIsUnique) {
	    # In "single key" files, comment out any extra lines
	    $line = $this->commentOut($key, $desc);
	}
	$seen->{$compKey}=1;
	$desc = $map->{$compKey}->[0];
      }

      # Print out dependancies
      my $depFilter = $this->getDependancies($compKey, $desc);
      if(defined $depFilter) {
	$this->closeBlock($map, $list, $seen, \@out, $depFilter);
      }

      push(@out, $line);
      next;
    }

    push(@out, $line);
  }

  # closes all open blocks, up to root. Necessary for files such as
  # smb.conf, which do not explicitely close their blocks...
  while(defined $map) {
    $this->closeBlock($map, $list, $seen, \@out);
    $map = pop(@mapStack);
    $list = pop(@listStack);
    $seen = pop(@seenStack);
  }
  writeFile($filename, \@out);
}

##
# Create an ConfigPatch object
#
sub new {
  my ($type) = @_;

  my $class = ref($type) || $type || "ConfigPatch";
  my $this= {};

  bless $this, $class;
}

# Read substitution file. Substitution file has (almost) the same syntax as 
# the config file to be patched.
# A substitution is a (key, value) pair.
# If the key is not present in the target config file, the key-value
# pair will be added to the end.
# If the key is present in the target config file, its value will be
# substituted
# TODO: support hierarchical substitutions
#
# Parameters:
# $this - this object.
# $file - the substitution file
#
# Return: the substitution map
# in addition, the substitution map is also stored statically in this object
#
# TODO: support multiline substitutions
sub readSubst {
  my ($this, $file) = @_;
  my $accum='';

  my @mapStack = ();
  my @listStack = ();
  my @keyStack = ();
  my @blockStack = ();

  my $map={};
  my $list=[];
  my $curKey;
  my $curBlock;

  ConfigSubs::init();
  $this->{map} = $map;
  $this->{list} = $list;

  open(IN, "<$file");
  $main::lineNo=0;
  while(<IN>) {
    my $line = $_;
    $main::lineNo++;

    if($line =~ /^@@%/) {
      # Skip intro line
      next;
    }

    $line = ConfigSubs::replaceVars($line);

    # is this a block start ?
    my ($key, $desc) = $this->isBlockStart($line);
    if(defined $key) {
      if(defined $curKey && $this->blockAutoEnd) {
	$map = pop(@mapStack);
	$list = pop(@listStack);
	$curKey = pop(@keyStack);
	$curBlock = pop(@blockStack);
      }
      push(@mapStack, $map);
      push(@listStack, $list);
      push(@keyStack, $curKey);
      push(@blockStack, $curBlock);

      my $newMap = { };
      $list = [];
      $curBlock = $map->{'B'.$key} = [ $desc, $newMap, $list, $accum ];
      $map = $newMap;
      $curKey = $key;
      next;
    }

    # is this a block end?
    my $closeDesc;
    ($key, $closeDesc) = $this->isBlockEnd($line, $curKey);
    if(defined $key) {
      if($key ne "" && $key ne $curKey) {
	die "[Patt] Misbalanced beginBlock ($curKey) and endBlock ($key) for $line\n";
      }
      if(!defined $curKey) {
	die "[ReadSubst] Attempt to close the root block\n";
      }
      $curBlock->[4] = $closeDesc;
      $map = pop(@mapStack);
      $list = pop(@listStack);
      $curKey = pop(@keyStack);
      $curBlock = pop(@blockStack);
    }

    ($key, $desc) = $this->isSetting(1, $line);
    if(defined $desc) {
      $map->{'S'.$key} = [ $desc, $accum ];
      push(@$list, [ $key, $desc, $accum]);
      $accum = "";
      next;
    }

    my $comm = $this->isComment($line);
    if(defined $comm) {
      $accum .= $comm;
      next;
    }

    $accum = "";
  }

  $this->{map};
}

##
# Customization: the following abstract methods need to be supplied
# for each type of config file syntax:
#  isComment   - Determines whether a line is a comment (called both on pattern
#		and target)
#  isSetting   - Determines whether a line is a setting (called both on pattern
#		and target)
#  subsSetting - Given a key substitutes value in a line (called on target)
#  isBlockStart - start of a block
#  isBlockEnd   - end of a block
#  printBlockStart - prints start of block
#  printBlockEnd - prints start of block


##
# Answers whether this is a comment
# Parameters
# $line - the line which is a comment or not
# Returns:
# The input line if it is indeed a comment, else undef
sub isComment {
  die "IsComment is an Abstract method\n";
}


##
# Determines whether this line is a setting
#
# Parameters
# $context  1 if parsing substitution pattern
#           0 if parsing config file to be modified
#           ---> can be used if extended syntax is to be implemented in
#           pattern file
# $line     the input line
#
# Returns:
#  - an opaque key identifying the setting
#  - an opaque  descriptor, that will later we passed the same way to
#    subsSetting
# or
#  (undef, undef) if this is not a setting
sub isSetting {
  die "IsSetting is an Abstract method\n";
}

##
# Substitutes setting
#
# Parameters:
# $this     - this object
# $key      - opaque substitution key
# $orig     - original line (in file to be patched)
# $subst    - the opaque substitution descriptor, as returned by isSetting
# $comments - any comments preceding the substitution (added for new insertion
#             to be re-output for new settings
# Returns:
#  The newly substituted line
sub subsSetting {
  my ($this, $key, $orig, $subst, $comments) = @_;
  if(!defined $subst) {
    printStackTrace;
    croak("Subst not defined\n");
  }
  my ($disabled, $origKey, $value) = (@$subst);
  $disabled.$origKey.$value;
}

##
# Determines whether this is the start of a block
# Parameters
# $this
# $line       - the line to analyze
#
# Returns:
#  (undef,undef)	if not start of a block
#  (blockId,desc)	if start of a block. BlockId is a unique identifier
#			for the block (within its parent block), whereas
#			desc is additional, non-identifying data
sub isBlockStart {
  return undef;
}

##
# Determines whether this is the end of a block
# Parameters
# $this
# $line       - the line to analyze
# $blockId    - the opaque identifier of the block: allows to detect bad
#             nesting
#
# Returns:
#  key if end
#  undef otherwise
sub isBlockEnd {
  return undef;
}

##
# Prints the start of a block (used for situations where blocks
# were in the pattern file but not in the target)
#
# Parameters
# $this
# $blockId
# $orig		- line before substitution
# $desc
# Returns:
#  the string representing the block start
sub printBlockStart {
  my ($this, $blockId, $orig, $desc) = @_;
  $desc;
}

##
# Prints the end of a block (used for situations where blocks
# were in the pattern file but not in the target)
#
# Parameters
# $this
# $blockId
# $orig		- line before substitution
# $desc
# Returns:
#  the string representing the block end
sub printBlockEnd {
    my ($this,$blockId,$orig,$desc,$closeDesc) = @_;
    $closeDesc;
}

##
# Block auto end: if true, blocks automatically end at the start of the next
# block
sub blockAutoEnd {
  0;
}

1;
