package ConfigSubs;

use English;
use strict;
use StackTrace;


# external interface
my ($mainIp, $mainMask);

# internal interface
my ($internalIp, $internalMask, $internalOrder);


my ($sec, $min, $hour, $mday, $mon, $year);
my $serial;

my $isInitialized;
my $netIsInitialized;
my $ldapIsInitialized;

my $hostname;
my $domainname;
my $idomainname;
my $orgdomainname;
my $llldomainname;
my $ypdomain;

my $baseDn;
my $usersDn;
my $groupsDn;

sub ipStringToNumber {
    my ($string) = @_;

    my (@ip, $accum);

    $accum = 0;
    foreach my $i (split(/\./, $string)) {
	$accum = $accum * 256 + $i;
    }
    return $accum;
}

sub ipNumberToString {
    my ($number) = @_;

    my ($i, @ip);
    @ip=();

    for($i=0; $i<4; $i++) {
	unshift(@ip, $number % 256);
	$number = int($number / 256);
    }
    return join(".", @ip);   
}

sub revIp {
    my ($number, $order) = @_;

    my @ip=();
    for(my $i=0; $i<4; $i++) {
	push(@ip, $number % 256);
	$number = int($number / 256);
    }
    return (join(".", @ip[4 - $order .. 3]), join(".", @ip[0 .. 3 - $order]));
}

sub shortIp {
    my ($number, $order) = @_;
    my @ip=();
    for(my $i=0; $i<4; $i++) {
	unshift(@ip, $number % 256);
	$number = int($number / 256);
    }

    return join(".", @ip[0 .. $order - 1]);
}

# the order is the number of complete bytes in the netmask
sub maskToOrder {
    my ($mask) = @_;
    my $order;

    if ($mask >= 0xffffffff) {
	$order = 4;
    } elsif ($mask >= 0xffffff00) {
	$order = 3;
    } elsif ($mask >= 0xffff0000) {
	$order = 2;
    } elsif ($mask >= 0xff000000) {
	$order = 1;
    } else {
	$order = 0;
    }
    return $order;
}

sub maskToBits {
  my ($mask) = @_;
  my $m=0;
  for(my $i=0; $i<33; $i++) {
    if($m == $mask) {
      return $i;
    }
    $m = ($m >> 1) | 0x80000000;
  }
}


# TODO: replace this with debian-compatible parser...
sub parseNetConfig {
  my ($file) = @_;

  open(NET_CONFIG, "<$file") ||
    die "Could not open $file ($ERRNO)\n";

  my %config;
  while(<NET_CONFIG>) {
    chop;
    if(/^([A-Za-z0-9_]*)=\"([^\"]*)\"$/) {
      $config{$1} = $2;
    } elsif(/^([A-Za-z0-9_]*)=\'([^\']*)\'$/) {
      $config{$1} = $2;
    } elsif(/^([A-Za-z0-9_]*)=(.*)$/) {
      $config{$1} = $2;
    }
  }
  close(NET_CONFIG);

  return digestNetConfig(\%config);
}

sub digestNetConfig {
  my ($config) = @_;

  my $ipString = $config->{IPADDR};
  my $ipNumber = ipStringToNumber($ipString);
  my $netmaskString = $config->{NETMASK};
  if($netmaskString == "") {
    $netmaskString = "255.255.255.255"; # Point-to-point
  }
  my $netmaskNumber = ipStringToNumber($netmaskString);
  my $subnet = ipNumberToString($netmaskNumber & $ipNumber);
  my $bcastString = $config->{BROADCAST};
  my $order = maskToOrder($netmaskNumber);
  my $bits = maskToBits($netmaskNumber);
  if($order == 4) {
    $order = 3;
  }
  my ($revIp, $ipLo) = revIp($ipNumber, $order);
  my $shortIp = shortIp($ipNumber, $order);

  return { ip => $ipString,
	   netmask => $netmaskString,
	   subnet => $subnet,
	   numNetmask => $netmaskNumber,
	   order => $order,
	   bits => $bits,
	   broadcast => $bcastString,
	   revIp => $revIp,
	   ipLo => $ipLo,
	   shortIp => $shortIp,
	   name => $config->{name}};
}

sub readLllInst {
  my ($hostNameP, $domainNameP, $fqHostNameP) = @_;
  open(LLLINST, "</etc/lllinst.cfg") ||
    open(LLLINST, "</mnt/etc/lllinst.cfg") ||
      return 0;
  while(<LLLINST>) {
    chomp;
    if(/^DOMAINNAME="(.*)"/) {
	$$domainNameP = $1;
    } elsif(/^FQHOSTNAME="(.*)"/) {
	$$fqHostNameP = $1;
    } elsif(/^HOSTNAME="(.*)"/) {
	$$hostNameP = $1;
    }

  }
  close(LLLINST);
  return 1;
}

my %netConfig;

# Goodness table
#   1 present in /etc/network/interfaces
#   2 present in ifconfig
#   4 has an IP address

# Gather info from ifconfig
#  - if we have info in /etc/network/interfaces, that info takes precedence
#  - in any case, we credit each interface from /etc/network/interfaces with
#    a goodness of 2
sub runIfconfig {
    open(IF, "/sbin/ifconfig -a |") || return;
    my $curIf;
    while(<IF>) {
	chomp;
	if(/^(\S+)/) {
	    $curIf={name=>$1};
	    if(! exists $netConfig{$1}) {
		$netConfig{$1}=$curIf;
	    } else {
		$curIf=$netConfig{$1};
	    }
	    $curIf->{goodness} |= 2;
	} elsif(/inet addr:(\S+)\s+(Bcast:(\S+)\s+)?Mask:(\S+)/) {
	  if(!defined $curIf->{IPADDR}) {
	    $curIf->{goodness} |= 4;
	    $curIf->{IPADDR}=$1;
	    $curIf->{BROADCAST}=$3;
	    $curIf->{NETMASK}=$4;
	  }
	} elsif(/inet addr:(\S+)\s+/) {
	  if(!defined $curIf->{IPADDR}) {
	    $curIf->{IPADDR}=$1;
	    $curIf->{NETMASK}="255.255.255.255";
	  }
	}
    }
    close(IF);

    open(ROUTE, "/sbin/route -n |") || return;
    while(<ROUTE>) {
	chomp;
	if(/^0\.0\.0\.0\s+(\S+)\s+0\.0\.0\.0.*\s(\S+)$/) {
	    if(exists $netConfig{$2}) {
		$netConfig{$2}->{GATEWAY}=$1;
		$netConfig{EXT}=$netConfig{$2};
	    }
	}
    }
    close(ROUTE);
}

sub readInterfaces {
    open(IF, "</etc/network/interfaces") || return;
    my $curIf;
    while(<IF>) {
	if(/^\s*iface\s*(\S+)\s+inet\s+static/) {
	    $curIf={ goodness => 1 };
	    $netConfig{$1}=$curIf;
	    $curIf->{name}=$1;
	} elsif(defined $curIf && /^\s*address\s+(\S+)/) {
	    $curIf->{IPADDR}=$1;
	    $curIf->{ goodness } |= 4;
	} elsif(defined $curIf && /^\s*gateway\s+(\S+)/) {
	    $curIf->{GATEWAY}=$1;
	    $netConfig{EXT}=$curIf;
	} elsif(defined $curIf && /^\s*(netmask|network|broadcast)\s+(\S+)/) {
	    $curIf->{uc($1)}=$2;
	} elsif(/^\s*auto/) {
	    $curIf=undef;
	}
    }
    close(IF);

    runIfconfig();

    my $goodness = 0;
    foreach my $value (values %netConfig) {
	if($value->{name} ne "lo" && !exists $value->{GATEWAY} &&
	    $value->{goodness} >= $goodness) {
	    $netConfig{INT}=$value;
	    $goodness = $value->{goodness};
	}
    }
}

sub readYpFile() {
    open(YP, "</etc/defaultdomain") || return;
    $ypdomain=<YP>;
    chomp($ypdomain);
    close(YP);
}

sub init {
  if($isInitialized) {
    return;
  }

  my @networks=();
  my $root=$ENV{ROOT};

  my $fqhostname;
  if(!readLllInst(\$hostname, \$domainname, \$fqhostname)) {
    open(HOSTNAME, "<$root/etc/hostname") || 
      die "Could not read hostname file ($ERRNO)\n";
    $hostname = <HOSTNAME>;
    chomp($hostname);
    close(HOSTNAME);

    if($hostname eq "") {
      die "Empty or missing hostname";
    }

    if($hostname !~ /^([^.]*)\.(.*)/) {
# OK, no dot in hostname. Check /etc/hosts for real domain name
      open(HOST, "</etc/hosts") || die "Could not read /etc/hosts ($ERRNO)\n";
      while(<HOST>) {
	chomp;
	s/\#.*//;
	s/^\s+//;
	my @line = split(/\s+/);
	my $found=0;
	foreach my $el (@line) {
	  if($el eq $hostname) {
	    $found=1;
	  } elsif($el =~ /^([^.]*)\.(.*)/ &&
		  $1 eq $hostname &&
		  $2 ne "") {
	    $domainname=$2;
	  }
	}
	last if($found);
      }
    } else {
      $hostname = $1;
      $domainname = $2;
    }
    if($domainname eq "") {
      print STDERR "Bad hostname $fqhostname\n";
      $domainname = "example.lll.lu";
    }
  }

  readYpFile();
  if(!defined $ypdomain) {
      $ypdomain=`ypdomainname`;
  }
  chomp($ypdomain);
  $idomainname = $domainname;
  $idomainname =~ s/\.([^.]*)$/\.intern/g;

  $orgdomainname = $domainname;
  $orgdomainname =~ s/\.([^.]*)$/\.org\.$1/g;

  $llldomainname = $domainname;
  $llldomainname =~ s/\.([^.]*)$/\.lll\.$1/g;

  ($sec, $min, $hour, $mday, $mon, $year) = localtime(time);
  $serial = sprintf "%04d%02d%02d02", ($year + 1900), $mon+1, $mday;
  $isInitialized = 1;
}

my $intNet;
my $extNet;

sub netInit {
  if($netIsInitialized) {
    return;
  }

  # foreach my $file (glob "$root/etc/sysconfig/network/ifcfg-eth*") {
  #   my $lnet = parseNetConfig("$file");
  #   if($lnet->{ip} =~ /^10\./ || $lnet->{ip} =~ /^192\.168/) {
  #     if(defined $intNet) {
  #	if($intNet->{numNetmask} < $lnet->{numNetmask}) {
  #	  # Intnet already exists, and is a larger net (less masklen)
  #	  # Leave alone

  #	  # The reason why we keep the larger net is in order to cope
  #	  # with a situation where we have a miniature external net to
  #	  # communicate with an ADSL modem

  #	  push(@ethNets, $lnet);
  #	} else {
  #	  push(@ethNets, $intNet);
  #	  $intNet=$lnet;
  #	}
  #     } else {
  #	$intNet = $lnet;
  #     }
  #   } else {
  #     # Does not match address pattern
  #     push(@ethNets, $lnet);
  #   }
  # }

  readInterfaces;

  my $intNetProto = $netConfig{'INT'};
  if(!defined $intNetProto) {
      $intNetProto = { };
  }

  if(!defined $intNetProto->{name}) {
      $intNetProto->{name} = "lo:0";
  }

  if(!defined $intNetProto->{IPADDR}) {
      $intNetProto->{IPADDR} = '10.0.0.1';
      $intNetProto->{NETMASK} = '255.0.0.0';
      $intNetProto->{BROADCAST} = '10.255.255.255';
  }

  $intNet = digestNetConfig($intNetProto);

  if(!defined $intNet) {
    die "Missing internal network\n";
  }
  print "Internal net: ".$intNet->{ip}."/". $intNet->{netmask}."\n";

  my $extIf;

  # TODO Make it configurable...

  my $extNetProto = $netConfig{'EXT'};

  if(!defined $extNetProto) {
      $extNetProto =  { IPADDR=>'158.64.137.230',
			NETMASK=>'255.255.255.0',
			BROADCAST=>'158.64.137.255' };
  }
  $extNet = digestNetConfig($extNetProto);

  print "External net: ".
    $extNet->{ip}."/". $extNet->{netmask}."\n";

  $netIsInitialized = 1;
}

sub ldapInit {
  if($ldapIsInitialized) {
    return;
  }

  open(SLAPCAT, "slapcat |");
  while(<SLAPCAT>) {
      chomp;
      if(/^dn: (dc=.*)$/) {
	  if(!defined $baseDn) {
	      $baseDn = $1;
	  }
      }
      if(/^dn: ou=(People|Users|Group|Groups),(dc=.*)$/) {
	  if($1 eq "People" || $1 eq "Users") {
	      $usersDn = $1;
	  }
	  if($1 eq "Group" || $1 eq "Groups") {
	      $groupsDn = $1;
	  }
      }
  }
  close(SLAPCAT);
  if(!defined $baseDn) {
    my $hn = `hostname -f`;
    chomp($hn);
    my @host = split(/\./,$hn);
    shift(@host);
    $baseDn = join(",", map { "dc=".$_ } @host);
  }
  if(!defined $usersDn) {
    $usersDn = "Users";
  }
  if(!defined $groupsDn) {
    $groupsDn = "Groups";
  }

  $ldapIsInitialized = 1;
}

sub domainname {
  init();
  return $domainname;
}

sub idomainname {
  init();
  return $idomainname;
}

sub orgdomainname {
  init();
  return $orgdomainname;
}

sub llldomainname {
  init();
  return $llldomainname;
}

sub hostname {
  init();
  return $hostname;
}

sub serial {
  init();
  return $serial;
}

sub ypdomain {
  init();
  return $ypdomain;
}


sub mainIpString {
  netInit();
  return $extNet->{ip};
}

sub mainNetmask {
  netInit();
  return $extNet->{netmask};
}

sub mainNmBits {
  netInit();
  return $extNet->{bits};
}

sub mainSubnet {
  netInit();
  return $extNet->{subnet};
}

sub mainBroadcast {
  netInit();
  return $extNet->{mainBroadcast};
}

sub mainIfname {
  netInit();
  return $extNet->{name};
}


sub intIpString {
  netInit();

  return $intNet->{ip};
}

sub intNetmask {
  netInit();
  return $intNet->{netmask};
}

sub intNmBits {
  netInit();
  return $intNet->{bits};
}

sub intSubnet {  
  netInit();
  if(!@_) {
      return $intNet->{subnet};
  } else {
      my $offset=$_[0];
      my $out=$intNet->{subnet};
      $out =~ s/\.[0-9]*$/.$offset/g; # Should really do binary calculation...
      return $out;
  }
}

sub intBroadcast {
  netInit();
  return $intNet->{broadcast};
}

sub intIfname {
  netInit();

  return $intNet->{name};
}

sub revIpV {
  netInit();
  return $extNet->{revIp};
}

sub irevIp {
  netInit();
  return $intNet->{revIp};
}


sub shortIpV {
  netInit();
  return $extNet->{shortIp};
}

sub ishortIp {
  netInit();
  return $intNet->{shortIp};
}

sub ipLo {
  netInit();
  return $extNet->{ipLo};
}

sub iipLo {
  netInit();
  return $intNet->{ipLo};
}

sub baseDn {
  ldapInit();
  return $baseDn;
}

sub usersDn {
  ldapInit();
  return $usersDn;
}

sub groupsDn {
  ldapInit();
  return $groupsDn;
}

my $ldapXPass;
sub ldapXPass {
    if(!defined $ldapXPass) {
	open(LDAPXPASS, "</root/ldapxpass.txt") ||
	    die "Could not read ldapxpass file\n";
	$ldapXPass = <LDAPXPASS>;
	chomp($ldapXPass);
	close(LDAPXPASS);
    }
    return $ldapXPass;
}

my $ldapPass;
sub ldapPass {
    if(!defined $ldapPass) {
	open(LDAPPASS, "</root/ldappass.txt") ||
	    die "Could not read ldappass file\n";
	$ldapPass = <LDAPPASS>;
	chomp($ldapPass);
	close(LDAPPASS);
    }
    return $ldapPass;
}

# This abstract class is used to patch config files. It should be
# subclassed with a class determining the exact syntax used by the
# config file to be patched.

sub replaceVars {
  my ($v) = @_;

  $v =~ s/\@\@DOMAIN\@\@/domainname()/ge;
  $v =~ s/\@\@IDOMAIN\@\@/idomainname()/ge;
  $v =~ s/\@\@ORGDOMAIN\@\@/orgdomainname()/ge;
  $v =~ s/\@\@LLLDOMAIN\@\@/llldomainname()/ge;

  $v =~ s/\@\@HOST\@\@/hostname()/ge;
  $v =~ s/\@\@SERIAL\@\@/serial()/ge;

  $v =~ s/\@\@YPDOMAIN\@\@/ypdomain()/ge;

  $v =~ s/\@\@MAINIP\@\@/mainIpString()/ge;
  $v =~ s/\@\@MAINNETMASK\@\@/mainNetmask()/ge;
  $v =~ s/\@\@MAINNMBITS\@\@/mainNmBits()/ge;
  $v =~ s/\@\@MAINSUBNET\@\@/mainSubnet()/ge;
  $v =~ s/\@\@MAINBROADCAST\@\@/mainBroadcast()/ge;
  $v =~ s/\@\@MAINIFNAME\@\@/mainIfname()/ge;

  $v =~ s/\@\@EXTIP\@\@/mainIpString()/ge;
  $v =~ s/\@\@EXTNETMASK\@\@/mainNetmask()/ge;
  $v =~ s/\@\@EXTNMBITS\@\@/mainNmBits()/ge;
  $v =~ s/\@\@EXTSUBNET\@\@/mainSubnet()/ge;
  $v =~ s/\@\@EXTBROADCAST\@\@/mainBroadcast()/ge;
  $v =~ s/\@\@EXTIFNAME\@\@/mainIfname()/ge;

  $v =~ s/\@\@INTIP\@\@/intIpString()/ge;
  $v =~ s/\@\@INTNETMASK\@\@/intNetmask()/ge;
  $v =~ s/\@\@INTNMBITS\@\@/intNmBits()/ge;
  $v =~ s/\@\@INTSUBNET\@\@/intSubnet()/ge;
  $v =~ s/\@\@INTSUBNET\[(\d+)\]\@\@/intSubnet($1)/ge;
  $v =~ s/\@\@INTBROADCAST\@\@/intBroadcast()/ge;
  $v =~ s/\@\@INTIFNAME\@\@/intIfname()/ge;

  $v =~ s/\@\@REVIP\@\@/revIpV()/ge;
  $v =~ s/\@\@IREVIP\@\@/irevIp()/ge;

  $v =~ s/\@\@SHORTIP\@\@/shortIpV()/ge;
  $v =~ s/\@\@ISHORTIP\@\@/ishortIp()/ge;

  $v =~ s/\@\@IPLO\@\@/ipLo()/ge;
  $v =~ s/\@\@IIPLO\@\@/iipLo()/ge;

  $v =~ s/\@\@BASEDN\@\@/baseDn()/ge;
  $v =~ s/\@\@USERSDN\@\@/usersDn()/ge;
  $v =~ s/\@\@GROUPSDN\@\@/groupsDn()/ge;

  $v =~ s/\@\@LDAPPASS\@\@/ldapPass()/ge;
  $v =~ s/\@\@LDAPXPASS\@\@/ldapXPass()/ge;

  if($v =~ /\@\@(.*?)\@\@/) {
      die "Bad key $1\n";
  }

  return $v;
}


# Test parsing of config
sub test {
  init();

  print "domainname=$domainname\n";
  print "idomainname=$idomainname\n";
  print "orgdomainname=$orgdomainname\n";
  print "llldomainname=$llldomainname\n";

  print "hostname=$hostname\n";
  print "mainIpString=".mainIpString()."\n";
  print "intIpString=".intIpString()."\n";

  print "revIp=".revIpV()."\n";
  print "irevIp=".iRevIp()."\n";

  print "shortIp=".shortIpV()."\n";
  print "ishortIp=".ishortIp()."\n";

  print "ipLo=".ipLo()."\n";
  print "iipLo=".iipLo()."\n";

  print "serial=$serial\n";
}

sub getDomain {
  init();
  return $domainname;
}

sub getHostname {
  init();
  return $hostname;
}

1;
