package MailMc;

use strict;
use ConfigPatch;
use Carp;
use StackTrace;

use vars qw/@ISA/;

@ISA=qw(ConfigPatch);

sub new {
  my ($type, @rest) = @_;
  my($class) = ref($type) || $type || "MailMc";

  my $this = $class->SUPER::new();
  bless $this, $class;

  if(@_ >= 2) {
    $this->readSubst($_[1]);
  }

  return $this;
}

##
# Answers whether this is a comment
# Parameters
# $line - the line which is a comment or not
# Returns:
# The input line if it is indeed a comment, else undef
sub isComment {
  my ($this, $line) = @_;
  if($line =~ /^(\s*dnl)/) {
    return $line;
  } else {
    return undef;
  }
}

##
# Determines whether this line is a setting
#
# Parameters
# $context  1 if parsing substitution pattern
#           0 if parsing config file to be modified
# $line     the input line
#
# Returns:
#  - an opaque key identifying the setting
#  - an opaque  descriptor, that will later we passed the same way to
#    subsSetting
#
# In this particular case, the key has the following format:
#  define.setting   = a define(`setting') line
#  feature.name     = a FEATURE(`name', `value') line
#  base.xyz         = a XYZ(`value') line
#
# The returned descriptor is as follows
#   [ isEnabled, value, trailer ]
sub isSetting {
  my ($this, $context, $line) = @_;

  if($line =~ /^(dnl\s*)?define\(\`([a-zA-Z0-9._]+)\'(.*)\)(.*)/) {
    return ("define.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?FEATURE\(\`dnsbl\', (\`[a-zA-Z0-9._]+\')(.*)\)(.*)/) {
    return ("dnsbl.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?FEATURE\(\`dnsbl\', ([a-zA-Z0-9._]+)(.*)\)(.*)/) {
    return ("dnsbl.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?FEATURE\((\`[a-zA-Z0-9._]+\')(.*)\)(.*)/) {
    return ("feature.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?FEATURE\(([a-zA-Z0-9._]+)(.*)\)(.*)/) {
    return ("feature.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?HACK\((\`[a-zA-Z0-9._]+\')(.*)\)(.*)/) {
    return ("hack.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?HACK\(([a-zA-Z0-9._]+)(.*)\)(.*)/) {
    return ("hack.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?include\((\`[a-zA-Z0-9._\/]+\')(.*)\)(.*)/) {
    return ("include.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?include\(([a-zA-Z0-9._\/]+)(.*)\)(.*)/) {
    return ("include.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?MAILER\((\`[a-zA-Z0-9._]+\')(.*)\)(.*)/) {
    return ("mailer.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s*)?MAILER\(([a-zA-Z0-9._]+)(.*)\)(.*)/) {
    return ("mailer.".$2, [ $1, $3, $4]);
  } elsif($line =~ /^(dnl\s+)?\s*([A-Z_]+)\((.*)\)(.*)/) {
    return ("bare.".$2, [ $1, $3, $4]);
  } else {
    return (undef, undef);
  }
}

# We depend on feature
sub dependFeature {
  my ($key) = @_;
  if($key =~ /^Sfeature\..*/) {
    return 1;
  } else {
    return 0;
  }
}


# Get dependancies
# Undef means that there are no dependancies
sub getDependancies {
  my ($this, $key) = @_;

  if($key =~ /^Smailer\..*/) {
    return \&dependFeature;
  } else {
    return undef;
  }
}


##
# Substitutes setting
#
# Parameters:
# $this     - this object
# $key      - opaque substitution key
# $orig     - original line (in file to be patched)
# $subst    - the opaque substitution descriptor, as returned by isSetting
# $comments - any comments preceding the substitution (added for new insertion
#             to be re-output for new settings
# Returns:
#  The newly substituted line
sub subsSetting {
  my ($this, $key, $orig, $subst, $comments) = @_;
  if(!defined $subst) {
    printStackTrace;
    croak("Subst not defined\n");
  }
#  print "Subst=".join(":", @$subst)."\n";
  my ($disabled, $value, $trailer) = (@$subst);
  my $middle;
  if($key =~ /define\.(.*)/) {
    $middle="define(`".$1."'";
  } elsif($key =~ /feature\.(.*)/) {
    $middle="FEATURE(".$1;
  } elsif($key =~ /hack\.(.*)/) {
    $middle="HACK(".$1;
  } elsif($key =~ /include\.(.*)/) {
    $middle="include(".$1;
  } elsif($key =~ /dnsbl\.(.*)/) {
    $middle="FEATURE(`dnsbl', ".$1;
  } elsif($key =~ /mailer\.(.*)/) {
    $middle="MAILER(".$1;
  } elsif($key =~ /bare\.(.*)/) {
    $middle=$1."(";
  } else {
    die "Bad key $key\n";
  }
  my $r = $disabled.$middle.$value.")".$trailer."\n";
  if(defined $orig) {
    return $r;
  } else {
    return $comments.$r;
  }
}


1;
