package NamedConf;

use strict;
use ConfigPatch;
use Carp;
use StackTrace;

use vars qw/@ISA/;

@ISA=qw(ConfigPatch);

sub new {
  my ($type, @rest) = @_;
  my($class) = ref($type) || $type || "NamedConf";

  my $this = $class->SUPER::new();
  bless $this, $class;

  if(@_ >= 2) {
    $this->readSubst($_[1]);
  }

  return $this;
}

##
# Answers whether this is a comment
# Parameters
# $line - the line which is a comment or not
# Returns:
# The input line if it is indeed a comment, else undef
sub isComment {
  my ($this, $line) = @_;
  if($line =~ /^(\s*;)/) {
    return $line;
  } else {
    return undef;
  }
}

##
# Determines whether this line is a setting
#
# Parameters
# $context  1 if parsing substitution pattern
#           0 if parsing config file to be modified
# $line     the input line
#
# Returns:
#  - an opaque key identifying the setting
#  - an opaque  descriptor, that will later we passed the same way to
#    subsSetting
#
# In this particular case, the key has the following format:
#
# The returned descriptor is as follows
#   [ isEnabled, value, trailer ]
sub isSetting {
  my ($this, $context, $line) = @_;

  if($line =~ /^(\s*\#)?(\s*([a-zA-Z0-9._-]+))(.*)/s) {
    return ($3, [ $1, $2, $4]);
  } else {
#    print "==> No\n";
    return (undef, undef);
  }
}

##
# Substitutes setting
#
# Parameters:
# $this     - this object
# $key      - opaque substitution key
# $orig     - original line (in file to be patched)
# $subst    - the opaque substitution descriptor, as returned by isSetting
# $comments - any comments preceding the substitution (added for new insertion
#             to be re-output for new settings
# Returns:
#  The newly substituted line
sub subsSetting {
  my ($this, $key, $orig, $subst, $comment) = @_;
  if(!defined $subst) {
    printStackTrace;
    croak("Subst not defined\n");
  }
  my ($disabled, $origKey, $value) = (@$subst);
  "$disabled$origKey$value";
}

##
# Determines whether this is the start of a block
# Parameters
# $this
# $line       - the line to analyze
#
# Returns:
#  (undef,undef)	if not start of a block
#  (blockId,desc)	if start of a block. BlockId is a unique identifier
#			for the block (within its parent block), whereas
#			desc is additional, non-identifying data
sub isBlockStart {
  my ($this, $line) = @_;

  if($line =~ /^\s*options\s*\{/) {
    ("options", undef);
  } elsif($line =~ /^\s*zone\s*\"(.*)\"\s*in\s*\{/i) {
    ("zone".$1, undef);
  } else {
    (undef, undef);
  }
}


##
# Determines whether this is the end of a block
# Parameters
# $this
# $line       - the line to analyze
# $blockId    - the opaque identifier of the block: allows to detect bad
#             nesting
#
# Returns:
#  key if end
#  undef otherwise
sub isBlockEnd {
  my ($this, $line, $blockId) = @_;
  if($line =~ /^\s*\}\s*;/) {
    ("",$line);
  } else {
    undef;
  }
}


##
# Prints the start of a block (used for situations where blocks
# were in the pattern file but not in the target)
#
# Parameters
# $this
# $blockId
# $orig		- line before substitution
# $desc
# Returns:
#  the string representing the block start
sub printBlockStart {
  my ($this, $blockId, $orig, $desc) = @_;

  if($blockId eq "options") {
    "options {\n";
  } elsif($blockId =~ /zone(.*)/) {
    "zone \"$1\" in {\n";
  } else {
    die "Bad blockId \"$blockId\"\n";
  }
}

##
# Prints the end of a block (used for situations where blocks
# were in the pattern file but not in the target)
#
# Parameters
# $this
# $blockId
# $orig		- line before substitution
# $desc
# Returns:
#  the string representing the block end
sub printBlockEnd {
  # nothing to do
  "};\n";
}

##
# Block auto end: if true, blocks automatically end at the start of the next
# block
sub blockAutoEnd {
  0;
}

1;
