package RcConfig;

use strict;
use ConfigPatch;
use Carp;
use StackTrace;

use vars qw/@ISA/;

@ISA=qw(ConfigPatch);

sub new {
  my ($type, @rest) = @_;
  my($class) = ref($type) || $type || "RcConfig";

  my $this = $class->SUPER::new();
  bless $this, $class;

  if(@_ >= 2) {
    $this->readSubst($_[1]);
  }

  return $this;
}

sub keyIsUnique {
    return 1;
}

##
# Answers whether this is a comment
# Parameters
# $line - the line which is a comment or not
# Returns:
# The input line if it is indeed a comment, else undef
sub isComment {
  my ($this, $line) = @_;
  if($line =~ /^(\s*\#)/) {
    return $line;
  } else {
    return undef;
  }
}

# Comments the line out
sub commentOut {
  my ($this, $key, $orig) = @_;

  my ($origDisabled, $origKey, $origValue, $origMode, $origSpc) = (@$orig);
  if($origDisabled ne "#") {
      $origDisabled = "# ";
  }
  return $origSpc.$origDisabled.$origKey.$origValue;
}

##
# Substitutes setting
#
# Parameters:
# $this     - this object
# $key      - opaque substitution key
# $orig     - triplet representing original line (in file to be patched)
# $subst    - the opaque substitution descriptor, as returned by isSetting
# $comments - any comments preceding the substitution (added for new insertion
#             to be re-output for new settings
# Returns:
#  The newly substituted line
sub subsSetting {
  my ($this, $key, $orig, $subst, $comments) = @_;
  if(!defined $subst) {
    printStackTrace;
    croak("Subst not defined\n");
  }
  my ($disabled, $zzkey, $value, $mode, $spc) = (@$subst);
  if(!defined $orig) {
      $orig=$subst;
  }
  my ($origDisabled, $origKey, $origValue, $origMode, $origSpc) = (@$orig);

  if($mode eq "LIST") {
    # List mode

    # Quick hack to leave commented-out lines alone
    if($orig->[0] =~ /^\#/) {

    }

    my $oDelim="";

    my $oValue = substr($orig->[2],1); # Strip off leading "=" sign
    $value=substr($value,1);
    chomp($value);
    chomp($oValue);

    if($oValue =~ /^"(.*)"/) {
      $oDelim="\"";
      $oValue=$1;
    }

    if($value =~ /^"(.*\S)"/) {
      $value=$1;
    }

    my @value = split(/\s+/, $value);

    # Make inventory of items to be removed
    my %removeList; # Elements to be removed
    foreach my $v (@value) {
      if($v =~ /^-(.*)/) {
	$removeList{$1}=1;
      }
    }

    # Add items which were in original line, minus those to be removed
    my %seen;     # Elements that are already present in original list
                    # don't add them twice
    my @resultList; # Result
    foreach my $v (split(/\s+/, $oValue)) {
      next if($removeList{$v});
      $seen{$v} = 1;
      push(@resultList, $v);
    }

    # Add items from substitution
    foreach my $v (@value) {
      next if($seen{$v});
      next if($v =~ /^-/);
      $seen{$v} = 1;
      push(@resultList, $v);
    }
    $value = "=".$oDelim.join(" ", @resultList).$oDelim."\n";
  }
  return $origSpc.$disabled.$key.$value;
}

##
# Determines whether this line is a setting
#
# Parameters
# $context  1 if parsing substitution pattern
#           0 if parsing config file to be modified
# $line     the input line
#
# Returns:
#  - an opaque key identifying the setting
#  - an opaque  descriptor, that will later we passed the same way to
#    subsSetting
#
# In this particular case, the key has the following format:
#  name             = name of the rc.config key
#
# The returned descriptor is as follows
#   [ isEnabled, value ]
sub isSetting {
  my ($this, $context, $line) = @_;

  if($line =~ /^(\s*)(\#\s*)?\[LIST\]\s*([a-zA-Z0-9._]+)(\s*=.*)/s) {
    return ($3, [ $2, $3, $4, 'LIST', $1]);
  } elsif($line =~ /^(\s*)(\#\s*)?([a-zA-Z0-9._]+)(\s*=.*)/s) {
    return ($3, [ $2, $3, $4, undef, $1]);
  } else {
    return (undef, undef);
  }
}

1;
