package SquidConf;

use strict;
use ConfigPatch;
use Carp;
use StackTrace;

use vars qw/@ISA/;

@ISA=qw(ConfigPatch);

sub new {
  my ($type, @rest) = @_;
  my($class) = ref($type) || $type || "SquidConf";

  my $this = $class->SUPER::new();
  bless $this, $class;

  if(@_ >= 2) {
    $this->readSubst($_[1]);
  }

  return $this;
}

##
# Answers whether this is a comment
# Parameters
# $line - the line which is a comment or not
# Returns:
# The input line if it is indeed a comment, else undef
sub isComment {
  my ($this, $line) = @_;
  if($line =~ /^(\s*\#)/) {
    return $line;
  } else {
    return undef;
  }
}

##
# Determines whether this line is a setting
#
# Parameters
# $context  1 if parsing substitution pattern
#           0 if parsing config file to be modified
# $line     the input line
#
# Returns:
#  - an opaque key identifying the setting
#  - an opaque  descriptor, that will later we passed the same way to
#    subsSetting
#
# In this particular case, the key has the following format:
#  define.setting   = a define(`setting') line
#  feature.name     = a FEATURE(`name', `value') line
#  base.xyz         = a XYZ(`value') line
#
# The returned descriptor is as follows
#   [ isEnabled, value, trailer ]
sub isSetting {
  my ($this, $context, $line) = @_;

  if($line =~ /^(\#? ?)((acl|deny_info|external_acl_type)\s+([a-zA-Z0-9._]+)\s+)(.*)/si) {
    return ("$3 $4", [ $1, $2, $5]);
  } elsif($line =~ /^(\#? ?)(http_access.*)/si) {
    return ("$2", [ $1, $2, ""]);
  } elsif($line =~ /^(\#)? ?(auth_param)\s+(\S+)\s+(\S+)(\s+.*)/si) {
    return ("$2 $3 $4", [ $1, "$2 $3 $4", $5]);
  } elsif($line =~ /^(\#)? ?([a-zA-Z0-9._]+)(\s+.*)/si) {
    return ($2, [ $1, $2, $3]);
  } else {
    return (undef, undef);
  }
}


# Get dependancies
# Undef means that there are no dependancies
# Given a key, returns a sub that says whether that sub's parameter should
# come before the key
sub getDependancies {
  my ($this, $key, $desc) = @_;
  my $comment = $desc->[0];
  if($comment =~ /\#/) {
    return undef; # Comment lines do not trigger any dependancies!
  }

  if($key =~ /^Shttp_access\s+(deny|allow)\s+all/) {
    # http_access deny/allow all must have all other http_access rules
    # precede, or else these will be ignored
    my $mode=$1;
    return sub {
      my ($key) = @_;
      if($key =~ /^Shttp_access\s+$mode\s+all/) {	
	return 0; # Universal allows/denies do not depend on themselves!
      }
      return $key =~ /^Shttp_access/;
    }
  } elsif($key =~ /^Shttp_access\s+(deny|allow)\s+(.*)/) {
    # acls should be defined before any http_access rule defining them
    my %acls;
    my $aclLine=$2;
    foreach my $acl ( split(/!|\s+!?/, $aclLine)) {
      if($acl ne "") {
	$acls{$acl} = 1;
      }
    }
    return sub {
      my ($key) = @_;
      if($key =~ /^Sacl\s+(\S+)/) {
	my $acl = $1;
	return exists $acls{$acl};
      }
      return 0;
    }
  } elsif($key =~ /^Sacl\s+/ && $desc->[2] =~ /^\s*proxy_auth/) {
    # auth_param should come before any proxy_auth ACL's
    return sub {
      my ($key) = @_;
      return $key =~ /^Sauth_param/;
    }
  } elsif($key =~ /^Sacl\s+/ && $desc->[2] =~ /^\s*external\s+(\S+)/) {
    # external_acl_types should come before any external acls using them
    my $neededType=$1;
    return sub {
      my ($key) = @_;
      if($key =~ /^Sexternal_acl_type\s+(\S+)/) {
	return $1 eq $neededType;
      }
      return 0;
    }
  } else {
    return undef;
  }
}


1;

